# encoding: utf-8
"""
groom

Created by Frodo on 10/26/19.
Copyright (c) 2019 Frodo. All rights reserved.
"""

import sys
import unittest
from . import *

__version__ = '0.0.3a3'

def function(
    *,
    first_arg: positional(int, 'int value', required=True, var_name='INT'),
    second_arg: switch('switch value', short_name='s'),
    third_arg: optional(str, 'string value', short_name='t') = 'string',
    fourth_arg: multiple(float, 'list of float values', short_name='f') = [0.1, 3.2]):
  print('first_arg:', first_arg)
  print('second_arg:', second_arg)
  print('third_arg:', third_arg)
  print('fourth_arg:', fourth_arg)

class DispatcherTest(unittest.TestCase):
  def setUp(self):
    self.disp = Dispatcher(
      function,
      "A test command-line program.")

  def test_helpmsg(self):
    sys.argv = [sys.argv[0], '-h']
    with self.assertRaises(SystemExit):
      self.disp.dispatch()

  def test_version(self):
    sys.argv = [sys.argv[0], '-v']
    with self.assertRaises(SystemExit):
      self.disp.dispatch()

  def test_positional(self):
    sys.argv = [sys.argv[0], '2']
    self.disp.dispatch()

  def test_required_positional_not_specify(self):
    sys.argv = [sys.argv[0]]
    with self.assertRaises(SystemExit):
      self.disp.dispatch()

  def test_keyword_switch(self):
    sys.argv = [sys.argv[0], '2', '--second-arg']
    self.disp.dispatch()

  def test_keyword_switch_short(self):
    sys.argv = [sys.argv[0], '2', '-s']
    self.disp.dispatch()

  def test_multiple_argument(self):
    sys.argv = [sys.argv[0], '2', '-f', '0.2', '-f', '0.5']
    self.disp.dispatch()

  def test_invaild_argument(self):
    sys.argv = [sys.argv[0], '--invalid']
    with self.assertRaises(SystemExit):
      self.disp.dispatch()

unittest.main()



# encoding: utf-8
# pylint: disable = redefined-outer-name, redefined-builtin
"""
annotations.py

Created by Frodo on 10/26/19.
Copyright (c) 2019 Frodo. All rights reserved.
"""

class Annotation:
  """
  class to annotate function parameters
  """
  def __init__(
      self, type_, desc='', *,
      allow_multiple=False,
      positional=False,
      short_name=None,
      required=False,
      var_name=None):
    if not isinstance(type_, type):
      raise TypeError(f"'{type(type_).__name__}' class is not inherits type class.")
    if type_ not in (str, int, float, complex, bool):
      raise TypeError(f"'{type_.__name__}' is not applicable type.")
    if type_ is bool and positional:
      raise TypeError("switch arg can't be a positional arg.")
    self.type = type_
    self.desc = desc
    self.allow_multiple = allow_multiple
    self.positional = positional
    self.short_name = short_name
    self.required = required
    self.var_name = var_name

  def display(self, arg: str, name: str):
    """
    generate human-readable parameter expression
    """
    disp = []
    if self.type is not bool:
      vn = self.var_name or name.upper()
      disp.append(f"{arg} <{vn}>")
      if self.short_name:
        disp.append(f" | -{self.short_name} <{vn}>")
      if self.allow_multiple:
        if disp[1:]:
          disp.insert(1, f" [{arg} <{vn}>...]")
          disp.append(f" [-{self.short_name} <{vn}>...]")
        else:
          disp.append(f" [{arg} <{vn}>...]")
      if not self.required:
        disp.insert(0, "[")
        disp.append("]")
    else:
      disp.append(f"{arg}")
      if self.short_name:
        disp.append(f" | -{self.short_name}")
        disp.insert(0, "(")
        disp.append(")")
    return "".join(disp)

def positional(type, desc='', *, required=False, var_name=None):
  """
  Annotation for positional arguments.
  """
  return Annotation(
    type, desc,
    positional=True,
    required=required,
    var_name=var_name)

def optional(type, desc='', *, var_name=None, short_name=None):
  """
  Annotation for keyword-only optional arguments.
  """
  return Annotation(
    type, desc,
    var_name=var_name,
    short_name=short_name)

def multiple(type, desc='', *, required=False, var_name=None, short_name=None):
  """
  Annotation for keyword-only list arguments.
  """
  return Annotation(
    type, desc,
    allow_multiple=True,
    required=required,
    var_name=var_name,
    short_name=short_name)

def required(type, desc='', *, var_name=None, short_name=None):
  """
  Annotation for keyword-only required arguments.
  """
  return Annotation(
    type, desc,
    required=True,
    var_name=var_name,
    short_name=short_name)

def switch(desc='', *, short_name=None):
  """
  Annotation for switch arguments.
  """
  return Annotation(
    bool, desc, short_name=short_name)



""""""
#  setup.py
#
#  Created by Frodo on 6/10/2019
#  Copyright (c) 2019 Frodo.

from os.path import dirname, abspath
from setuptools import setup, find_packages

try:
  with open("README.rst") as f:
    readme = f.read()
except IOError:
  readme = ""

here = dirname(abspath(__file__))
version = '0.0.3a3'

setup(
  name="groom",
  version=version,
  url="https://github.com/frodo821/groom",
  author="frodo821 <Twitter: @BoufrawFrodo2>",
  author_email='1234567890.sakai.jp@gmail.com',
  maintainer="frodo821 <Twitter: @BoufrawFrodo2>",
  maintainer_email='1234567890.sakai.jp@gmail.com',
  description="A easy-to-use command-line interface generator",
  long_description=readme,
  packages=find_packages(),
  install_requires=[],
  license="MIT",
  classifiers=[
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'License :: OSI Approved :: MIT License',
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Operating System :: OS Independent',
    'Topic :: Utilities',
    'Topic :: Terminals',
    'Topic :: Software Development :: Libraries :: Application Frameworks'
  ],
  entry_points="")



# encoding: utf-8
"""
groom

Created by Frodo on 10/26/19.
Copyright (c) 2019 Frodo. All rights reserved.
"""

from .annotations import (
  positional, optional,
  multiple, required, switch)
from .dispatching import Dispatcher

__all__ = [
  "positional", "optional",
  "multiple", "required",
  "switch", "Dispatcher"
]

__version__ = "0.0.3a1"



# encoding: utf-8
"""
dispatching.py

Created by Frodo on 10/26/19.
Copyright (c) 2019 Frodo. All rights reserved.
"""

import sys
from os.path import basename
from inspect import signature as sig, Parameter
from typing import List, Dict, Tuple, Union

from .annotations import Annotation

__all__ = ["Dispatcher"]

_Param = Union[str, float, int, complex, bool]

def _to_param_name(name):
  return f"--{name.lower().replace('_', '-')}"

def _get_program_name():
  from __main__ import __package__ as pkg
  p = pkg or sys.argv[0]
  return basename(p)

def _generate_disp_info(func) -> Tuple[Dict[str, _Param], Dict[str, Annotation], List[Annotation]]:
  defaults, keywords, positionals = {}, {}, []
  for n, p in sig(func).parameters.items():
    ann = p.annotation
    if ann is Parameter.empty:
      raise ValueError(f"parameter '{n}' is not annotated.")
    if not isinstance(ann, Annotation):
      raise TypeError(f"unexpected annotation type: {type(ann).__name__}")
    defaults[n] = p.default if p.default is not Parameter.empty else None
    if ann.positional:
      positionals.append((n, ann))
      continue
    keywords[_to_param_name(n)] = n, ann
    if ann.short_name:
      keywords[f"-{ann.short_name}"] = n, ann
  positionals.sort(key=lambda x: not x[1].required)
  return defaults, keywords, positionals

class Dispatcher:
  """
  Dispatches command-line call to python function call.

  Parameters:

  * func: a function to call
  * desc: tool description
  """
  positionals: List[Annotation]
  keywords: Dict[str, Annotation]
  defaults: Dict[str, _Param]

  def __init__(self, func=None, desc: str = None, *, is_subcommand: bool = False):
    self.func = func
    self.desc = desc
    self.is_subcommand = is_subcommand
    self.subdisps = {}
    if self.func is None:
      return
    self.defaults, self.keywords, self.positionals = _generate_disp_info(func)

  def __check_default_action(self, arg):
    if arg in ('-h', '--help'):
      print(self.helpmsg())
      sys.exit(0)
    if arg in ('-v', '--version'):
      import __main__ as m
      print(f"{_get_program_name()}: {getattr(m, '__version__', 'UNVERSIONED')}")
      sys.exit(0)

  def __handle_not_keyword(self, arg, idx, pos, params):
    if idx == 1 and self.subdisps:
      sd = self.subdisps.get(arg)
      if sd is None:
        print((f"unexpected subcommand: '{arg}'\n"
               "please try execute this command with '-h' or '--help' to get helps."),
              file=sys.stderr)
        sys.exit(-1)
      sys.argv.pop(0)
      sd.dispatch()
      sys.exit(0)
    try:
      n, p = next(pos)
    except StopIteration:
      print((f"unexpected argument: '{arg}'\n"
             "please try execute this command with '-h' or '--help' to get helps."),
            file=sys.stderr)
      sys.exit(-1)
    try:
      params[n] = p.type(arg)
    except ValueError:
      print((f"invalid literal '{arg}' for type '{p.type.__name__}'\n"
             "please try execute this command with '-h' or '--help' to get helps."),
            file=sys.stderr)
      exit(-1)

  def __post_validate(self, params):
    for name, param in self.positionals:
      if name not in params:
        if param.required:
          print((f"some of required positional parameter was not specified.\n"
                 "please try execute this command with '-h' or '--help' to get helps."),
                file=sys.stderr)
          sys.exit(-1)
        params[name] = self.defaults[name]
    for name, param in self.keywords.values():
      if name not in params:
        if param.required:
          print((f"required parameter '{_to_param_name(name)}' was not specified.\n"
                 "please try execute this command with '-h' or '--help' to get helps."),
                file=sys.stderr)
          sys.exit(-1)
        if param.type is bool:
          params[name] = False
          continue
        params[name] = self.defaults[name]

  def dispatch(self):
    """
    Dispatches command-line call to python function call.
    """
    params = {}
    pos = iter(self.positionals)
    idx = 1
    while sys.argv[idx:]:
      arg = sys.argv[idx]
      self.__check_default_action(arg)
      if not arg.startswith('-'):
        self.__handle_not_keyword(arg, idx, pos, params)
        idx += 1
        continue
      try:
        n, p = self.keywords.get(arg)
      except TypeError:
        print((f"unexpected argument: '{arg}'\n"
               "please try execute this command with '-h' or '--help' to get helps."),
              file=sys.stderr)
        sys.exit(-1)
      if p.allow_multiple and n not in params:
        params[n] = []
      if p.type is bool:
        params[n] = True
        idx += 1
        continue
      try:
        v = p.type(sys.argv[idx+1])
      except IndexError:
        print((f"no value passed for parameter '{arg}'\n"
               "please try execute this command with '-h' or '--help' to get helps."),
              file=sys.stderr)
        sys.exit(-1)
      except ValueError:
        print(
          (f"invalid literal '{arg}' for type '{p.type.__name__}'\n"
           "please try execute this command with '-h' or '--help' to get helps."),
          file=sys.stderr)
        exit(-1)
      if p.allow_multiple:
        params[n].append(v)
      else:
        params[n] = v
      idx += 2
    self.__post_validate(params)
    self.func(**params)

  def __generate_usage(self, name):
    ret = []
    ps = []
    for n, p in self.positionals:
      ps.append(f"<{p.var_name}>" if p.required else f"[{p.var_name or n.upper()}]")
    ret.append(' '.join(ps))
    for arg, (n, p) in filter(lambda x: x[0].startswith('--'), self.keywords.items()):
      ret.append(f"    {p.display(arg, n)}")
    args = '\n'.join(ret).strip()
    return f"  {name} {args}"

  def helpmsg(self):
    """
    generate help message
    """
    import __main__ as m
    pn = _get_program_name()
    ret = [
      self.desc
    ] if self.is_subcommand else [
      f"{pn}: {getattr(m, '__version__', 'UNVERSIONED')}",
      "",
      self.desc,
      "",
      "Usage:",
      f"  {pn} [-v | --version | -h | --help]"
    ]
    if self.subdisps:
      if not self.is_subcommand:
        ret.append(f"  {pn} subcommand params...")
        ret.append("")
      ret.append("subcommands:")
      for sn, sc in self.subdisps.items():
        ret.append(f"{sn}:")
        ret.append(sc.helpmsg().replace('\n', '\n  '))
    else:
      ret.append(self.__generate_usage(pn))
      ret.append("")
      ret.append("positional parameters:")
      for n, p in self.positionals:
        ret.append(f"{p.var_name or n.upper()}:")
        ret.append(f"  {p.desc}")
        ret.append(f"  type: {p.type.__name__}")
        ret.append(f"  required: {p.required}")
        if not p.required:
          ret.append(f"  default: {self.defaults[n]}")
      ret.append("")
      ret.append("parameters:")
      for word, (n, p) in self.keywords.items():
        if not word.startswith('--'):
          continue
        ret.append(f"{word + (f', -{p.short_name}' if p.short_name else '')}:")
        ret.append(f"  {p.desc}")
        ret.append(f"  type: {p.type.__name__}")
        ret.append(f"  required: {p.required}")
        ret.append(f"  multiple values: {p.allow_multiple}")
        if not p.required:
          ret.append(f"  default: {self.defaults[n]}")
      return '\n'.join(ret)

  def add_subcommand(self, name, dispatcher):
    """
    Add a dispatcher as a sub-command.
    """
    dispatcher.is_subcommand = True
    self.subdisps[name] = dispatcher
